(* #load "camlp4o.cma";; *)
(* ocamlc -thread -custom unix.cma threads.cma getopt.cma -pp camlp4o -I +camlp4 -I getopt agitest.ml *)
(*  -cclib -lunix -cclib -lthreads *)
(* ocamlopt -thread  -pp camlp4o -I +camlp4 -I getopt unix.cmxa threads.cmxa getopt.cmxa agitest.ml *)
(* -cclib -lunix  -cclib -lthreadsnat  -cclib -lpthread *)

(* ./a.out -h "localhost" -p 4573 *)
(* For stack trace, compile with -g and "export OCAMLRUNPARM='b'" *)

open Genlex;;
open List;;
open Printf;;
(* open Thread;; *)
open Unix;;
open Array;;
open Getopt;;
open String;;
open Commands;; 
open Dirrecurse;;

let host = ref ""
and port = ref ""
and notimeout = ref false
and nohangup = ref false
and loadtest = ref ""
and startdir = ref "";;

let specs = [
  ('h', "host", None, (atmost_once host (Error "only one host")));
  ('p', "port", None, (atmost_once port (Error "only one port")));
  (' ', "notimeout", (Getopt.set notimeout true), None);
  (' ', "nohangup", (Getopt.set nohangup true), None);
  ('l', "loadtest", None, (atmost_once loadtest (Error "only one loadtest")));
  ('s', "startdir", None, (atmost_once startdir (Error "only one start directory")));
]

let _ = 
  parse_cmdline specs print_endline;

host := if !host = "" then "localhost" else !host;
port := if !port = "" then "4573" else !port;
loadtest := if !loadtest = "" then  "1" else !loadtest;;
startdir := if !startdir = "" then "." else !startdir;;

let string_of_list lst = 
  (concat "\n" (List.flatten lst))

let print_list lst = 
  printf "%s%!\n" (string_of_list lst)

(* Code under GPL from http://pauillac.inria.fr/~guesdon/Tools/share/code_Sfile.html *)
let string_of_file name =
  let chanin = open_in_bin name in
  let len = 1024 in
  let s = String.create len in
  let buf = Buffer.create len in
  let rec iter () =
    try
      let n = input chanin s 0 len in
      if n = 0 then
        ()
      else
        (
         Buffer.add_substring buf s 0 n;
         iter ()
        )
    with
      End_of_file -> ()
  in
  iter ();
  close_in chanin;
  Buffer.contents buf;;
(******************************************************************************)




(*
let lexer = make_lexer [":";"<";">"];;

let rec parse_command = parser
    [< 'Ident "STREAM"; 'Ident "FILE"; r = rest parse_streamfile >] -> r
  | [< 'Ident "GET"; r = get >] -> r
  | [< 'Ident "VERBOSE"; r = rest parse_verbose >] -> r
  | [< 'Ident "ANSWER"; r = rest  parse_answer >] -> r
  | [< 'Ident "AUTOHANGUP"; r = rest parse_autohangup >] -> r
  | [< 'Ident "CHANNEL"; 'Ident "STATUS"; r = rest parse_channelstatus >] -> r
  | [< 'Ident "EXEC"; r = rest parse_exec >] -> r
  | [< 'Ident "SET"; r = set >] -> r
  | [< 'Ident "HANGUP"; r = rest parse_hangup >] -> r
  | [< 'Ident "RECEIVE"; 'Ident "CHAR"; r = rest parse_receivechar >] -> r
  | [< 'Ident "RECORD"; 'Ident "FILE"; r = rest parse_recordfile >] -> r
  | [< 'Ident "SAY"; r = say >] -> r
  | [< 'Ident "SEND"; r = send >] -> r
  | [< 'Ident "TDD"; 'Ident "MODE"; r = rest parse_tddmode >] -> r
  | [< 'Ident "WAIT"; 'Ident "FOR"; 'Ident "DIGIT"; r = rest parse_waitfordigit >] -> r
  | [< 'Kwd "<"; 'String s; 'Kwd ">"; n = parse_command >] -> n
  | [< >] -> []
and get = parser
    [< 'Ident "VARIABLE"; r = rest parse_getdata >] -> r
  | [< 'Ident "DATA"; r = rest parse_getdata >] -> r
and say = parser
    [< 'Ident "DIGITS"; r = rest parse_saydigits >] -> r
  | [< 'Ident "NUMBER"; r = rest parse_saynumber >] -> r
and send = parser
    [< 'Ident "IMAGE"; r = rest parse_sendimage >] -> r
  | [< 'Ident "TEXT"; r = rest parse_sendtext >] -> r
and set = parser
    [< 'Ident "VARIABLE"; r = rest parse_setvar >] -> r
  | [< 'Ident "CALLERID"; r = rest parse_setcallerid >] -> r
  | [< 'Ident "CONTEXT"; r = rest parse_setcontext >] -> r
  | [< 'Ident "EXTENSION"; r = rest parse_setextension >] -> r
  | [< 'Ident "PRIORITY"; r = rest parse_setpriority >] -> r
and rest pfunc = parser
    [< p = pfunc; n = parse_command >] -> List.append [p] n
and parse_return ret = parser
    [< 'Kwd ":"; 'String keysp >] -> [sprintf "200 result=%s" keysp]
  | [< >] -> [sprintf "200 result=%s" ret]
and parse_streamfile = parser
    [< 'String p; 'String keys; n = parse_return "-1" >] -> ["SF"] @ n @ [p; keys]
and parse_verbose = parser
    [< 'Int lev >] -> ["V"; string_of_int lev]
and parse_answer = parser
    [< 'Kwd ":"; 'String ret >] -> ["A"; sprintf "200 result=%s" ret]
  | [< >] -> ["A"; "200 result=-1"]
and parse_getdata = parser
    [< 'String f; p = gd_params;  k = parse_return "-1" >] -> ["GD"] @ k @ [f] @ p
and gd_params = parser
    [< 'Int n; p = gd_params >] -> [string_of_int n] @ p
  | [< >] -> []
and parse_autohangup = parser
    [< 'Int time >] -> ["AH"; string_of_int time]
and parse_channelstatus = parser
    [< 'String ch; n = parse_return "-1" >] -> ["CS"; ch] @ n
and parse_exec = parser
    [< 'String app; 'String opts; n = parse_return "-2" >] -> ["E"; app; opts] @ n
and parse_getvar = parser
    [< 'String var >] -> ["GV"; var]
and parse_setvar = parser
    [< 'String var; 'String v >] -> ["SV"; var; v]
and parse_hangup = parser
    [< 'String ch; n = parse_return "-1" >] -> ["H"; ch] @ n
and parse_receivechar = parser
    [< 'Int timeout; n = parse_return "-1" >] -> ["RC"] @ n @ [string_of_int timeout]
and parse_recordfile = parser
    [< 'String f; 'String ext; 'String keys; 'Int timeout; 
       b = parse_recordfile ; 'Kwd ":"; 'String keysp >] -> 
      ["RF"; keysp; ext; keys; string_of_int timeout] @ b
  | [< 'Ident "beep" >] -> ["beep"]
  | [< >] -> []
and parse_saydigits = parser
    [< 'Int digits; 'String keys; n = parse_return "-1" >] -> ["SD"] @ n @[string_of_int digits; keys]
and parse_saynumber = parser
    [< 'Int digits; 'String keys; n = parse_return "-1" >] -> ["SN"] @ n @ [string_of_int digits; keys]
and parse_sendimage = parser
    [< 'String img; n = parse_return "-1" >] -> ["SI"] @ n @ [img]
and parse_sendtext = parser
    [< 'String txt; n = parse_return "-1" >] -> ["ST"] @ n @[ txt]
and parse_setcallerid = parser
    [< 'String name; 'Kwd "<"; 'Int num; 'Kwd ">" >] -> ["SCI"; name; string_of_int num]
  | [< 'Ident name; 'Kwd "<"; 'Int num; 'Kwd ">" >] -> ["SCI"; name; string_of_int num]
  | [< 'Kwd "<"; 'Int num; 'Kwd ">" >] -> ["SCI"; string_of_int num]
and parse_setcontext = parser
    [< 'String context >] -> ["SC"; context]
and parse_setextension = parser
    [< 'Int ext >] -> ["SE"; string_of_int ext]
and parse_setpriority = parser
    [< 'Int pri >] -> ["SP"; string_of_int pri]
and parse_tddmode = parser
    [< 'String setting; n = parse_return "-1" >] -> ["TM"] @ n @ [setting]
and parse_waitfordigit = parser
    [< 'Int timeout; n = parse_return "-1" >] -> ["WFD"] @ n @ [string_of_int timeout]

let parse_commands = parser [< n = parse_command; _ = Stream.empty >] -> n;;
*)


exception No_such_action;;
exception Action_mismatch of string;;
exception Incomplete of string;;

let commands_eq sact fact =
  if (List.length sact.params) = (List.length fact.params) then
    let mem_eq s f = match f with
      "_" -> true
    | "-999" -> true
    | str -> s = f in
    let all_true = List.map2 mem_eq sact.params fact.params in
    not (List.mem false all_true)
  else
    false;;

let vars = Hashtbl.create 10
let perform_action sact fact hangup =
  let check s f act hup =
    if (commands_eq s f) then
      if hup = None || f.hangup = "" then
          begin printf "%s %s\n" act f.hangup; act end
      else
        (concat "" ["200 result=";f.hangup])
    else
      raise (Action_mismatch (concat " " fact.name)) in
  if fact.abbrev <> sact.abbrev then 
    raise (Action_mismatch (concat " " ["from file:"; (concat " " fact.name); ", from server:"; (concat " " sact.name)]))
  else
    match fact.abbrev with
      "SF" | "GD" | "CS" | "E" | "H" | "RC" | "RF" | "SD" | "SN" | "SI" | "ST" | "TM" | "WFD" ->
        check sact fact fact.return hangup
    | "AH" | "SC" | "SE" | "SP" ->
        check sact fact "200 result=0" hangup
    | "V" | "SCI" ->
        check sact fact "200 result=-1" hangup
    | "GV" ->
        begin
          try
            let v = Hashtbl.find vars fact.return in
            check sact fact (sprintf "200 result=1 (%s)" v) hangup
          with Not_found -> "200 result=0"
        end
    | "SV" ->
        begin
          Hashtbl.add vars fact.return (hd sact.params);
          check sact fact "200 result=1" hangup
        end
    | "A" ->
       (hd sact.params)
    | r -> raise No_such_action;;


(*

let new_call commands = 
  print_list (List.flatten commands);;

let t = Thread.create new_call (read_file);;
t;;

let t = Thread.create new_call (read_command8);;
t;;


*)

let sock = Unix.socket PF_INET SOCK_STREAM 0;;
connect sock (ADDR_INET((gethostbyname !host).h_addr_list.(0), int_of_string !port));;

let bufsize = 512
let _ =
  let read_file f = Lexing.from_channel (open_in_bin f)
  and write_msg msg = 
    let n = ref 0 
    and msg = (concat "" [msg; "\n"]) in
    printf "msg %s%!" msg;
    while !n >= 0 && !n < (String.length msg) do
      let tmp = Unix.write sock msg (!n) ((String.length msg)-(!n)) in
      ignore (n := !n + if tmp > 0 then tmp else -1);
    done;
  and read_msg() =
    let n = ref 0 
    and  buf = String.make bufsize '\000' in
    while !n >= 0 && !n < bufsize do
      let tmp = Unix.read sock buf (!n) (bufsize-(!n)) in
      ignore (n := if tmp > 0 then (!n + tmp) else -2);
      ignore (n := if ((!n <> -2) && (String.get buf (!n - 1) = '\n')) then -1 else -2);
    done; 
    if !n <> -2 then buf else ""
  and all_com f =
    let com = ref [] in 
    (try
      while true do
        com := List.append !com [Agiparse.main Agilex.token f]
      done
    with Agilex .Eof -> ignore ());
    !com in
  let check_server f = 
    let sact m = (hd (all_com (Lexing.from_string m)))
    and msg = ref "" in
    let rec verify_server fact cnt =
      let ret = ref (Some 0)
      and cnt2 = ref cnt in
      if (List.length fact) > 0 then begin 
        msg := read_msg();
        if !msg <> "" then 
          begin
            begin
              let sactm = sact !msg in
              try
                if !cnt2 > 0 then
                  write_msg (perform_action sactm (hd fact) None)
                else
                  write_msg (perform_action sactm (hd fact) (Some 0));
                  if (hd fact).hangup <> "" then cnt2 := cnt - 1;
              with Action_mismatch s ->
                printf "Unexpected command from FAGI: %s\nFollows (server) (file)\n(%s)\n(%s)%!"
                  s (concat " " sactm.params) (concat " " (hd fact).params);
                (exit 0)
            end;
            if !cnt2 >= 0 then ignore (ret := verify_server (tl fact) !cnt2); 
          end
        else
          raise (Incomplete (sprintf "Expecting: %s\n" (concat " " (hd fact).name)));
      end else ret := None;
      !ret
    and inc = ref 0 in
    if !nohangup then 
      ignore (verify_server (all_com (read_file f)) 99999999)
    else 
      while (verify_server (all_com (read_file f)) !inc) <> None  do
        inc := !inc + 1
      done;
    ignore (Unix.system (concat "/" [!startdir;"cleanup"])) in
  let action files abstr = match (hd files) with
    "agivars" -> write_msg (string_of_file "agivars")
  | "init" -> check_server "init"
  | s -> printf "--%s\n" s; if (String.sub s ((String.length s) - 5) 5) = ".test" then 
      check_server s in
  recur "." action ["agivars"; "init"] [] []
